/*
 * Copyright (C) 2005-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */


/** \file afs_common.c Functions common to all audio file selectors. */

#include "server.cmdline.h"
#include "server.h"
#include "vss.h"
#include <dirent.h> /* readdir() */
#include <sys/stat.h> /* stat */
#include <sys/types.h> /* mode_t */
#include "error.h"
#include "string.h"

/**
 * Traverse the given directory recursively.
 *
 * \param dirname The directory to traverse.
 * \param f The function to call for each entry.
 *
 * For each regular file whose filename ends in .yyy, where yyy is a supported
 * audio format, the supplied function \a f is called.  The directory and
 * filename component of the regular file are passed to \a f.
 *
 * \return On success, 1 is returned. Otherwise, this function returns a
 * negative value which indicates the kind of the error.
 */
int find_audio_files(const char *dirname, int (*f)(const char *, const char *))
{
	DIR *dir = NULL;
	struct dirent *entry;
	/*
	 * Opening the current directory (".") and calling fchdir() to return
	 * is usually faster and more reliable than saving cwd in some buffer
	 * and calling chdir() afterwards (see man 3 getcwd).
	 */
	int cwd_fd = open(".", O_RDONLY);
	struct stat s;
	int ret;

	if (cwd_fd < 0)
		return -E_GETCWD;
	ret = -E_CHDIR;
	if (chdir(dirname) < 0)
		goto out;
	ret = -E_OPENDIR;
	dir = opendir(".");
	if (!dir)
		goto out;
	/* scan cwd recursively */
	while ((entry = readdir(dir))) {
		mode_t m;
		char *tmp;

		if (!strcmp(entry->d_name, "."))
			continue;
		if (!strcmp(entry->d_name, ".."))
			continue;
		ret = -E_LSTAT;
		if (lstat(entry->d_name, &s) == -1)
			continue;
		m = s.st_mode;
		if (!S_ISREG(m) && !S_ISDIR(m)) /* skip links, sockets, ... */
			continue;
		if (S_ISREG(m)) { /* regular file */
			if (guess_audio_format(entry->d_name) < 0)
				continue;
			ret = f(dirname, entry->d_name);
			if (ret < 0)
				goto out;
			continue;
		}
		/* directory */
		tmp = make_message("%s/%s", dirname, entry->d_name);
		ret = find_audio_files(tmp, f);
		free(tmp);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	if (dir)
		closedir(dir);
	if (fchdir(cwd_fd) < 0)
		ret = -E_CHDIR;
	close(cwd_fd);
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	return ret;
}
